import time

import tonio
import tonio.sync
import tonio.sync.channel as channel


def test_semaphore(run):
    counter = 0

    def _count(semaphore, i):
        nonlocal counter
        with (yield semaphore()):
            counter += 1
            if counter > 2:
                raise RuntimeError
            yield
            counter -= 1
        return i

    def _run(value):
        semaphore = tonio.sync.Semaphore(value)
        out = yield tonio.spawn(*[_count(semaphore, i) for i in range(50)])
        return out

    assert run(_run(2)) == list(range(50))

    # TODO: need global raise
    # with pytest.raises(RuntimeError):
    #     run(_run(3))


def test_lock(run):
    counter = 0

    def _count(semaphore, i):
        nonlocal counter
        with (yield semaphore()):
            counter += 1
            if counter > 1:
                raise RuntimeError
            yield
            counter -= 1
        return i

    def _run():
        semaphore = tonio.sync.Lock()
        out = yield tonio.spawn(*[_count(semaphore, i) for i in range(50)])
        return out

    assert run(_run()) == list(range(50))


def test_barrier(run):
    barrier = tonio.sync.Barrier(3)
    count = 0

    def _foo():
        nonlocal count
        count += 1
        i = yield barrier.wait()
        time.sleep(0.1)
        assert count == 3
        return i

    def _run():
        out = yield tonio.spawn(*[_foo() for _ in range(3)])
        return out

    assert set(run(_run())) == {0, 1, 2}


def test_channel(run):
    def _produce(sender, barrier, offset, no):
        for i in range(no):
            message = offset + i
            yield sender.send(message)
        yield barrier.wait()

    def _consume(receiver, target, count):
        messages = []
        while count < target:
            try:
                message = yield receiver.receive()
                count += 1
                messages.append(message)
            except Exception:
                break
        return messages

    def _close(sender, barrier):
        yield barrier.wait()
        sender.close()

    def _run2p4c():
        count = 0
        sender, receiver = channel.channel(2)
        barrier = tonio.sync.Barrier(3)
        tasks = [
            _produce(sender, barrier, 100, 20),
            _produce(sender, barrier, 200, 20),
            _consume(receiver, 40, count),
            _consume(receiver, 40, count),
            _consume(receiver, 40, count),
            _consume(receiver, 40, count),
            _close(sender, barrier),
        ]
        [_, _, c1, c2, c3, c4, _] = yield tonio.spawn(*tasks)
        return c1, c2, c3, c4

    def _run4p2c():
        count = 0
        sender, receiver = channel.channel(2)
        barrier = tonio.sync.Barrier(5)
        tasks = [
            _produce(sender, barrier, 100, 10),
            _produce(sender, barrier, 200, 10),
            _produce(sender, barrier, 300, 10),
            _produce(sender, barrier, 400, 10),
            _consume(receiver, 40, count),
            _consume(receiver, 40, count),
            _close(sender, barrier),
        ]
        [_, _, _, _, c1, c2, _] = yield tonio.spawn(*tasks)
        return c1, c2

    consumed = run(_run2p4c())
    consumed = {v for c in consumed for v in c}
    assert len(consumed) == 40
    assert consumed == ({*range(100, 120)} | {*range(200, 220)})

    consumed = run(_run4p2c())
    consumed = {v for c in consumed for v in c}
    assert len(consumed) == 40
    assert consumed == ({*range(100, 110)} | {*range(200, 210)} | {*range(300, 310)} | {*range(400, 410)})



from ._tonio import Event as _Event, Waiter as Waiter


class Event(_Event):
    def wait(self, timeout: int | float | None = None):
        timeout = round(max(0, timeout * 1_000_000)) if timeout else timeout
        yield self.waiter(timeout)

    async def __call__(self, timeout: int | float | None = None):
        timeout = round(max(0, timeout * 1_000_000)) if timeout else timeout
        await self.waiter(timeout)



from functools import wraps

from ._runtime import run


def main(
    *coros, threads: int | None = None, threads_blocking: int | None = None, threads_blocking_timeout: int | None = None
):
    if not coros:
        #: opts
        def deco(coro):
            @wraps(coro)
            def wrapper():
                run(
                    coro(),
                    threads=threads,
                    threads_blocking=threads_blocking,
                    threads_blocking_timeout=threads_blocking_timeout,
                )

            return wrapper

        return deco

    if len(coros) > 1:
        raise SyntaxError('Invalid argument for `main`')

    [coro] = coros

    @wraps(coro)
    def wrapper():
        run(coro())

    return wrapper



from socket import SocketType as _SocketType
from typing import Any, TypeVar

_T = TypeVar('_T')

__version__: str

class Runtime:
    _clock: int
    _closed: bool
    _sig_listening: bool
    _sig_loop_handled: bool
    _stopping: bool

    def __init__(self, threads: int, threads_blocking: int, threads_blocking_timeout: int): ...
    def _run(self): ...
    def _spawn_pygen(self, coro): ...
    def _spawn_pyasyncgen(self, coro): ...
    def _spawn_blocking(self, f, *args, **kwargs) -> tuple[Event, ResultHolder]: ...
    def reader_add(self, fd: int, event: Event): ...
    def reader_rem(self, fd: int): ...
    def writer_add(self, fd: int, event: Event): ...
    def writer_rem(self, fd: int): ...

class Waiter:
    def __init__(self, *events: Event): ...

class Event:
    def __init__(self): ...
    def set(self): ...
    def clear(self): ...
    def is_set(self) -> bool: ...
    def waiter(self, timeout: int | None) -> Waiter: ...

class ResultHolder:
    def __init__(self, size: int = 1): ...
    def store(self, value, index: int | None = None): ...
    def fetch(self): ...

class Lock:
    def acquire(self) -> None | Event: ...
    def release(self): ...

class LockCtx:
    def __init__(self, lock: Lock): ...
    def __enter__(self): ...
    def __exit__(self, exc_type, exc_value, exc_tb): ...

class Semaphore:
    def __init__(self, value: int): ...
    def acquire(self) -> None | Event: ...
    def release(self): ...

class SemaphoreCtx:
    def __init__(self, semaphore: Semaphore): ...
    def __enter__(self): ...
    def __exit__(self, exc_type, exc_value, exc_tb): ...

class Barrier:
    event: Event

    def __init__(self, value: int): ...
    def ack(self) -> int: ...

class Channel:
    def __init__(self, size: int): ...

class ChannelSender:
    def __init__(self, channel: Channel): ...
    def _send_or_wait(self, message: Any) -> None | Event: ...
    def _send(self, message): ...
    def close(self): ...

class ChannelReceiver[_T]:
    def __init__(self, channel: Channel): ...
    def _receive(self) -> tuple[_T, Event | None]: ...

class UnboundedChannel:
    def __init__(self): ...

class UnboundedChannelSender[_T]:
    def __init__(self, channel: UnboundedChannel): ...
    def send(self, message: _T): ...

class UnboundedChannelReceiver[_T]:
    def __init__(self, channel: UnboundedChannel): ...
    def _receive(self) -> tuple[_T, Event | None]: ...

class CancelledError(BaseException): ...
class RuntimeAlreadyInitializedError(RuntimeError): ...
class RuntimeNotInitializedError(RuntimeError): ...
class TimeoutError(BaseException): ...
class WouldBlock(Exception): ...

class Socket:
    _sock: _SocketType

    def __init__(self, stdlib_socket: _SocketType): ...

def get_runtime() -> Runtime: ...
def set_runtime(runtime: Runtime): ...



import contextlib
from typing import TypeVar

from ._events import Event
from ._tonio import ResultHolder, TimeoutError, get_runtime
from ._types import Coro


_T = TypeVar('_T')


def clock() -> int:
    return get_runtime()._clock


def sleep(timeout: int | float) -> Coro[None]:
    yield from Event().wait(timeout)


def timeout(coro: Coro[_T], timeout: int | float) -> Coro[tuple[None | _T, bool]]:
    done = Event()
    res = ResultHolder()

    def wrapper():
        ret = yield coro
        res.store(ret)
        done.set()

    get_runtime()._spawn_pygen(wrapper())

    yield from done.wait(timeout)
    if not done.is_set():
        with contextlib.suppress(TimeoutError):
            coro.throw(TimeoutError)
        return None, False
    return res.fetch(), True



import pytest

from tonio._runtime import Runtime
from tonio._tonio import set_runtime
from tonio._utils import is_asyncg


@pytest.fixture(scope='session')
def runtime():
    rv = Runtime(threads=4, threads_blocking=8, threads_blocking_timeout=10)
    set_runtime(rv)
    return rv


@pytest.fixture(scope='session')
def run(runtime):
    def inner(coro):
        runner = runtime.run_pyasyncgen_until_complete if is_asyncg(coro) else runtime.run_pygen_until_complete
        return runner(coro)

    return inner



from typing import Callable, ParamSpec, TypeVar

from ._events import Event, Waiter
from ._tonio import ResultHolder, get_runtime
from ._types import Coro


_Params = ParamSpec('_Params')
_Return = TypeVar('_Return')


def spawn(*coros: Coro):
    events = []
    res = ResultHolder(len(coros))

    def wrapper(idx, coro, event):
        ret = yield coro
        res.store(ret, idx)
        event.set()

    for idx, coro in enumerate(coros):
        event = Event()
        events.append(event)
        get_runtime()._spawn_pygen(wrapper(idx, coro, event))

    waiter = Waiter(*events)

    def join():
        yield waiter
        return res.fetch()

    return join()


def spawn_blocking(fn: Callable[_Params, _Return], /, *args: _Params.args, **kwargs: _Params.kwargs) -> Coro[_Return]:
    event, res = get_runtime()._spawn_blocking(fn, *args, **kwargs)
    yield event.waiter(None)
    return res.fetch()



import time

import tonio


def test_sleep(run):
    def f():
        start = time.monotonic()
        yield tonio.spawn(tonio.sleep(0.05), tonio.sleep(0.1))
        return time.monotonic() - start

    assert run(f()) >= 0.1


def _sleep(x):
    yield tonio.sleep(x)
    return 3


def _test_timeout():
    out1, success1 = yield tonio.timeout(_sleep(0.2), 0.3)
    out2, success2 = yield tonio.timeout(_sleep(0.2), 0.1)
    assert out1 == 3
    assert out2 is None
    assert success1 is True
    assert success2 is False


def test_timeout(run):
    run(_test_timeout())



from ._tonio import (
    CancelledError as CancelledError,
    RuntimeAlreadyInitializedError as RuntimeAlreadyInitializedError,
    RuntimeNotInitializedError as RuntimeNotInitializedError,
    TimeoutError as TimeoutError,
    WouldBlock as WouldBlock,
)



from __future__ import annotations

import os
import socket as _stdlib_socket
from types import TracebackType
from typing import Any

from .._ctl import spawn_blocking
from .._tonio import Event, Socket as _SocketWrapper, get_runtime
from .._types import Coro


class _Socket(_SocketWrapper):
    def detach(self) -> int:
        return self._sock.detach()

    def fileno(self) -> int:
        return self._sock.fileno()

    def getpeername(self) -> Any:
        return self._sock.getpeername()

    def getsockname(self) -> Any:
        return self._sock.getsockname()

    def getsockopt(
        self,
        level: int,
        optname: int,
        buflen: int | None = None,
    ) -> int | bytes:
        if buflen is None:
            return self._sock.getsockopt(level, optname)
        return self._sock.getsockopt(level, optname, buflen)

    def setsockopt(
        self,
        level: int,
        optname: int,
        value: Any,
        optlen: int | None = None,
    ) -> None:
        if optlen is None:
            if value is None:
                raise TypeError(
                    "invalid value for argument 'value', must not be None when specifying optlen",
                )
            return self._sock.setsockopt(level, optname, value)
        if value is not None:
            raise TypeError(
                f"invalid value for argument 'value': {value!r}, must be None when specifying optlen",
            )

        return self._sock.setsockopt(level, optname, value, optlen)

    def listen(self, backlog: int = min(_stdlib_socket.SOMAXCONN, 128)) -> None:
        return self._sock.listen(backlog)

    def get_inheritable(self) -> bool:
        return self._sock.get_inheritable()

    def set_inheritable(self, inheritable: bool) -> None:
        return self._sock.set_inheritable(inheritable)

    def __enter__(self) -> _Socket:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        return self._sock.__exit__(exc_type, exc_value, traceback)

    @property
    def family(self) -> Any:
        return self._sock.family

    @property
    def type(self) -> int:
        return self._sock.type

    @property
    def proto(self) -> int:
        return self._sock.proto

    # @property
    # def did_shutdown_SHUT_WR(self) -> bool:
    #     return self._did_shutdown_SHUT_WR

    def __repr__(self) -> str:
        return repr(self._sock).replace('socket.socket', 'tonio.net.socket.socket')

    def dup(self) -> _Socket:
        return _Socket(self._sock.dup())

    def close(self) -> None:
        if self._sock.fileno() != -1:
            # TODO: cleanup runtime refs
            # get_runtime().notify_closing(self._sock)
            self._sock.close()

    def bind(self, address: Any) -> Coro[None]:
        # TODO: error handling
        # print('BIND')
        address = yield self._resolve_address(address, local=True)
        # print('BIND', address)
        yield spawn_blocking(self._sock.bind, address)
        # print('BOUND')

    def shutdown(self, flag: int) -> None:
        self._sock.shutdown(flag)
        # if flag in [_stdlib_socket.SHUT_WR, _stdlib_socket.SHUT_RDWR]:
        #     self._did_shutdown_SHUT_WR = True

    def _resolve_address(
        self,
        address: Any,
        *,
        local: bool,
    ) -> Coro[Any]:
        # print('SRESADDR', self, address)
        if self.family == _stdlib_socket.AF_INET6:
            ipv6_v6only = self._sock.getsockopt(
                _stdlib_socket.IPPROTO_IPV6,
                _stdlib_socket.IPV6_V6ONLY,
            )
        else:
            ipv6_v6only = False
        addr = yield _resolve_address(
            self.type,
            self.family,
            self.proto,
            ipv6_v6only=ipv6_v6only,
            address=address,
            local=local,
        )
        return addr

    def accept(self):
        # print('SOCK ACCEPT')
        runtime = get_runtime()
        fd = self.fileno()
        event = Event()
        runtime.reader_add(fd, event)
        # print('SOCK ACCEPT added reader', self)

        while True:
            yield event.waiter(None)
            try:
                conn, address = self._sock.accept()
            except (BlockingIOError, InterruptedError):
                event.clear()
                continue
            except BaseException as exc:
                runtime.reader_rem(fd)
                raise exc
            else:
                runtime.reader_rem(fd)
                break

        # print('SOCK ACCEPT DONE', conn, address)
        return from_stdlib_socket(conn), address

    def connect(self, address: Any) -> Coro[None]:
        # print('SOCK CONNECT', address)
        address = yield self._resolve_address(address, local=False)
        # print('SOCK CONNECT ADDR', address)

        try:
            self._sock.connect(address)
        except (BlockingIOError, InterruptedError):
            pass
        else:
            return

        runtime = get_runtime()
        fd = self.fileno()
        event = Event()
        runtime.writer_add(fd, event)
        # print('SOCK CONNECT added writer', self)

        while True:
            yield event.waiter(None)
            try:
                err = self._sock.getsockopt(_stdlib_socket.SOL_SOCKET, _stdlib_socket.SO_ERROR)
                if err != 0:
                    raise OSError(err, 'Connect call failed %s' % (address,))
            except (BlockingIOError, InterruptedError):
                event.clear()
                continue
            except BaseException as exc:
                runtime.writer_rem(fd)
                raise exc
            else:
                runtime.writer_rem(fd)
                break

    def recv(self, bufsize: int, flags: int = 0, /) -> Coro[bytes]:
        # print('RECV', self)
        runtime = get_runtime()
        fd = self.fileno()
        event = Event()
        runtime.reader_add(fd, event)
        # print('RECV added reader', self)
        data = b''

        while True:
            yield event.waiter(None)
            try:
                data += self._sock.recv(bufsize, flags)
            except (BlockingIOError, InterruptedError):
                # print('RECV BLOCK EVTCLEAR', self)
                event.clear()
                continue
            except BaseException as exc:
                runtime.reader_rem(fd)
                raise exc
            else:
                runtime.reader_rem(fd)
                break

        # print('RECV DONE', self)
        return data

    # coro def recv_into
    # coro def recvfrom
    # coro def recvfrom_into
    # coro def recvmsg
    # coro def recvmsg_into

    def send(self, data: Any, flags: int = 0, /) -> Coro[int]:
        if not bytes:
            return 0

        try:
            n = self._sock.send(data, flags)
        except (BlockingIOError, InterruptedError):
            n = 0

        if n == len(data):
            return n

        runtime = get_runtime()
        fd = self.fileno()
        event = Event()
        runtime.writer_add(fd, event)
        sent = n

        while True:
            yield event.waiter(None)
            try:
                n = self._sock.send(data[sent:], flags)
            except (BlockingIOError, InterruptedError):
                event.clear()
                continue
            except BaseException as exc:
                runtime.writer_rem(fd)
                raise exc

            sent += n
            if sent == len(data):
                runtime.writer_rem(fd)
                break

        return sent

    # coro def sendto
    # coro def sendmsg

    # not impl
    # coro def sendfile

    # Intentionally omitted:
    #   sendall
    #   makefile
    #   setblocking/getblocking
    #   settimeout/gettimeout
    #   timeout


def from_stdlib_socket(sock: _stdlib_socket.socket) -> _Socket:
    return _Socket(sock)


def socket(
    family: int = _stdlib_socket.AF_INET,
    type: int = _stdlib_socket.SOCK_STREAM,
    proto: int = 0,
    fileno: int | None = None,
) -> _Socket:
    # TODO: handle fileno (get opts)
    stdlib_socket = _stdlib_socket.socket(family, type, proto, fileno)
    return from_stdlib_socket(stdlib_socket)


def getaddrinfo(
    host: bytes | str | None,
    port: bytes | str | int | None,
    family: int = 0,
    type: int = 0,
    proto: int = 0,
    flags: int = 0,
) -> Coro[
    list[
        tuple[
            Any,
            int,
            int,
            str,
            tuple[str, int] | tuple[str, int, int, int] | tuple[int, bytes],
        ]
    ]
]:
    # print('GETADDRINFO')
    ret = yield spawn_blocking(
        _stdlib_socket.getaddrinfo,
        host,
        port,
        family,
        type,
        proto,
        flags,
    )
    return ret


def _resolve_address(
    type_: int,
    family: Any,
    proto: int,
    *,
    ipv6_v6only: bool | int,
    address: Any,
    local: bool,
) -> Coro[Any]:
    # print('RESADDR', address)
    if family == _stdlib_socket.AF_INET:
        if not isinstance(address, tuple) or not len(address) == 2:
            raise ValueError('address should be a (host, port) tuple')
    elif family == _stdlib_socket.AF_INET6:
        if not isinstance(address, tuple) or not 2 <= len(address) <= 4:
            raise ValueError(
                'address should be a (host, port, [flowinfo, [scopeid]]) tuple',
            )
    elif hasattr(_stdlib_socket, 'AF_UNIX') and family == _stdlib_socket.AF_UNIX:
        assert isinstance(address, (str, bytes, os.PathLike))
        return os.fspath(address)
    else:
        return address

    host: str | None
    host, port, *_ = address
    if isinstance(port, int) and host is not None:
        try:
            _stdlib_socket.inet_pton(family, host)
        except (OSError, TypeError):
            pass
        else:
            return address

    if host == '':
        host = None
    if host == '<broadcast>':
        host = '255.255.255.255'
    flags = 0
    if local:
        flags |= _stdlib_socket.AI_PASSIVE
    if family == _stdlib_socket.AF_INET6 and not ipv6_v6only:
        flags |= _stdlib_socket.AI_V4MAPPED
    gai_res = yield getaddrinfo(host, port, family, type_, proto, flags)
    assert len(gai_res) >= 1
    (*_, normed), *_ = gai_res
    if family == _stdlib_socket.AF_INET6:
        list_normed = list(normed)
        assert len(normed) == 4
        if len(address) >= 3:
            list_normed[2] = address[2]
        if len(address) >= 4:
            list_normed[3] = address[3]
        return tuple(list_normed)
    return normed



import tonio
from tonio.net import socket


_SIZE = 1024 * 1024


def _recv_all(sock: socket.SocketType, nbytes):
    buf = b''
    while len(buf) < nbytes:
        buf += yield sock.recv(nbytes - len(buf))
    return buf


def test_socket_accept_recv(run):
    def server():
        sock = socket.socket()

        with sock:
            yield sock.bind(('127.0.0.1', 0))
            sock.listen()

            task = tonio.spawn(client(sock.getsockname()))

            client_sock, _ = yield sock.accept()
            with client_sock:
                data = yield _recv_all(client_sock, _SIZE)

            yield task

        return data

    def client(addr):
        sock = socket.socket()
        with sock:
            yield sock.connect(addr)
            yield sock.send(b'a' * _SIZE)

    data = run(server())
    assert data == b'a' * _SIZE


def test_socket_accept_send(run):
    state = {'data': b''}

    def server():
        sock = socket.socket()

        with sock:
            yield sock.bind(('127.0.0.1', 0))
            sock.listen()

            task = tonio.spawn(client(sock.getsockname()))

            client_sock, _ = yield sock.accept()
            with client_sock:
                yield client_sock.send(b'a' * _SIZE)

            yield task

    def client(addr):
        sock = socket.socket()
        with sock:
            yield sock.connect(addr)
            while len(state['data']) < _SIZE:
                state['data'] += yield sock.recv(1024 * 16)

    run(server())
    assert state['data'] == b'a' * _SIZE



import contextlib

from ._tonio import (
    Barrier as _Barrier,
    Channel as _Channel,
    ChannelReceiver as _ChannelReceiver,
    ChannelSender as _ChannelSender,
    Lock as _Lock,
    LockCtx as _LockCtx,
    Semaphore as _Semaphore,
    SemaphoreCtx as _SemaphoreCtx,
    UnboundedChannel as _UnboundedChannel,
    UnboundedChannelReceiver as _UnboundedChannelReceiver,
    UnboundedChannelSender as UnboundedChannelSender,
)
from ._types import Coro


class Lock(_Lock):
    def __call__(self) -> Coro[contextlib.AbstractContextManager[None]]:
        if event := self.acquire():
            yield event.waiter(None)
        return _LockCtx(self)

    async def __aenter__(self):
        if event := self.acquire():
            await event()

    async def __aexit__(self):
        self.release()


class Semaphore(_Semaphore):
    def __call__(self) -> Coro[contextlib.AbstractContextManager[None]]:
        if event := self.acquire():
            yield event.waiter(None)
        return _SemaphoreCtx(self)

    async def __aenter__(self):
        if event := self.acquire():
            await event()

    async def __aexit__(self):
        self.release()


class Barrier(_Barrier):
    def wait(self):
        count = self.ack()
        yield self.event.waiter(None)
        return count


class ChannelSender(_ChannelSender):
    def send(self, message) -> Coro[None]:
        if event := self._send_or_wait(message):
            yield event.waiter(None)
            self._send(message)


class ChannelReceiver(_ChannelReceiver):
    def receive(self) -> Coro[None]:
        msg, event = self._receive()
        while event:
            yield event.waiter(None)
            msg, event = self._receive()
        return msg


class UnboundedChannelReceiver(_UnboundedChannelReceiver):
    def receive(self) -> Coro[None]:
        msg, event = self._receive()
        while event:
            yield event.waiter(None)
            msg, event = self._receive()
        return msg


def channel(size) -> tuple[ChannelSender, ChannelReceiver]:
    inner = _Channel(size)
    return ChannelSender(inner), ChannelReceiver(inner)


def unbounded_channel() -> tuple[UnboundedChannelSender, UnboundedChannelReceiver]:
    inner = _UnboundedChannel()
    return UnboundedChannelSender(inner), UnboundedChannelReceiver(inner)



import multiprocessing

from ._events import Event
from ._tonio import ResultHolder, Runtime as _Runtime, set_runtime as _set_runtime
from ._utils import is_asyncg


class Runtime(_Runtime):
    def run_forever(self):
        try:
            self._run_forever_pre()
            self._run()
        finally:
            self._run_forever_post()

    def _run_forever_pre(self):
        # TODO: signals
        pass

    def _run_forever_post(self):
        # TODO: signals
        self._stopping = False

    def run_pygen_until_complete(self, coro):
        done = Event()
        res = ResultHolder()

        def wrapper():
            ret = yield coro
            res.store(ret)
            done.set()

        def watcher():
            yield from done.wait()
            self.stop()

        self._spawn_pygen(watcher())
        self._spawn_pygen(wrapper())
        self.run_forever()

        return res.fetch()

    def run_pyasyncgen_until_complete(self, coro):
        done = Event()
        res = ResultHolder()

        async def wrapper():
            ret = await coro
            res.store(ret)
            done.set()

        async def watcher():
            await done()
            self.stop()

        self._spawn_pyasyncgen(watcher())
        self._spawn_pyasyncgen(wrapper())
        self.run_forever()

        return res.fetch()

    def stop(self):
        self._stopping = True


def run(coro, **opts):
    # print(
    #     coro,
    #     inspect.isasyncgen(coro),
    #     inspect.isasyncgenfunction(coro),
    #     inspect.iscoroutine(coro),
    #     inspect.iscoroutinefunction(coro),
    # )
    opts['threads'] = opts.get('threads') or multiprocessing.cpu_count()
    opts['threads_blocking'] = opts.get('threads_blocking') or 128
    opts['threads_blocking_timeout'] = opts.get('threads_blocking_timeout') or 30
    runtime = Runtime(**opts)
    _set_runtime(runtime)
    runner = runtime.run_pyasyncgen_until_complete if is_asyncg(coro) else runtime.run_pygen_until_complete
    return runner(coro)
